import numpy as np
import pytest
from vito.cam_projections import dot, apply_transformation, \
    apply_projection, P_from_K_R_t, project_world_to_image_K_Rt, \
    project_world_to_image_K_R_t, project_world_to_image_K_R_C, \
    project_world_to_image_with_distortion_K_Rt, \
    project_world_to_image_with_distortion_K_R_t, \
    project_world_to_image_with_distortion_K_R_C, \
    get_groundplane_to_image_homography, shift_points_along_viewing_rays, \
    apply_dehomogenization, C_from_Rt, C_from_R_t, \
    get_image_to_groundplane_homography, \
    normalize_image_coordinates, normalize_image_coordinates_with_distortion, \
    compare_rotation_matrices, rot3d, transform, project


def test_dot():
    x = np.random.rand(3, 1)
    y = np.random.rand(3)
    d = x[0]*y[0] + x[1]*y[1] + x[2]*y[2]
    assert dot(x, y) == d
    assert dot(y, x) == d

    with pytest.raises(ValueError):
        dot(np.zeros((10)), np.zeros((20)))
    with pytest.raises(ValueError):
        dot(np.zeros((10,)), np.zeros((20,)))
    with pytest.raises(ValueError):
        dot(np.array([]), np.zeros((20,)))
    with pytest.raises(ValueError):
        dot(np.zeros((13,)), np.array([]))
    assert dot(np.float32(3), np.float64(7)) == 21
    assert dot(np.float32(3), np.float64([7])) == 21
    assert dot(np.float32([3]), np.float64([7])) == 21


def test_apply_transformation():
    # Ensure alias is set up
    assert transform == apply_transformation
    T2d = np.random.rand(2, 2)
    T3d = np.random.rand(3, 2)

    with pytest.raises(ValueError):
        x = apply_transformation(T2d, np.int32(2))
    x = apply_transformation(T2d, np.int32([17]))
    assert x.shape[0] == 2 and x.shape[1] == 1
    expected = 17 * T2d[:, 0] + T2d[:, 1]
    assert x[0, 0] == pytest.approx(expected[0])
    assert x[1, 0] == pytest.approx(expected[1])

    x = apply_transformation(T2d, np.ones((2, 1)))
    assert x.shape[0] == 2 and x.shape[1] == 1
    assert np.all(np.sum(T2d, axis=1).reshape(x.shape) == x)

    x = apply_transformation(T3d, 2*np.ones((2, 1), dtype=np.uint8))
    assert x.shape[0] == 3 and x.shape[1] == 1
    assert x.dtype == np.float64
    assert np.all(np.sum(T3d, axis=1).reshape(x.shape) == pytest.approx(x/2))

    # 3x2 * 1x1
    x = apply_transformation(T3d, 2*np.ones((1, 1), dtype=np.uint8))
    assert x.shape[0] == 3 and x.shape[1] == 1
    assert x.dtype == np.float64
    expected = 2*T3d[:, 0] + T3d[:, 1]
    assert np.all(expected.reshape(x.shape) == pytest.approx(x))

    with pytest.raises(ValueError):
        # 3x2 * 3x1
        apply_transformation(T3d, np.ones((3, 1), dtype=np.uint8))
    with pytest.raises(ValueError):
        # 3x2 * 2x2
        apply_transformation(np.random.rand(3, 4), np.random.rand(1, 2))


def test_apply_projection():
    # Ensure alias is set up
    assert project == apply_projection
    T2d = np.random.rand(2, 2)
    x = apply_projection(T2d, 2*np.ones((2, 3)))
    assert x.shape[0] == 1 and x.shape[1] == 3
    s = np.sum(T2d, axis=1)
    expected = s[0] / s[1]
    assert np.all(expected == pytest.approx(x))

    # Project the world origin to some point.
    # This works, because the homogeneous coordinate will be added.
    T = np.random.rand(3, 4)
    x = apply_projection(T, np.zeros((3, 1)))
    assert x.shape[0] == 2 and x.shape[1] == 1
    assert x[0] == pytest.approx(T[0, 3]/T[2, 3])
    assert x[1] == pytest.approx(T[1, 3]/T[2, 3])
    # 3D-2D projection can also be done via:
    y = project_world_to_image_K_Rt(np.eye(3, 3), T, np.zeros((3, 1)))
    assert np.all(y == x)
    # Now, try difference project_world_to_image functions:
    pts = np.random.rand(3, 17)
    K = np.diag(np.random.rand(3)) * 92
    K[2, 2] = 1
    R = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    t = np.random.rand(3, 1)
    Rt = np.column_stack((R, t))
    y = project_world_to_image_K_Rt(K, Rt, pts)
    z = project_world_to_image_K_R_t(K, R, t, pts)
    assert np.all(z == y)
    # Or even abusing the projection with(out) distortion
    no_distortion = project_world_to_image_with_distortion_K_Rt(
        K, Rt, np.zeros((5, 1)), pts)
    assert np.all(no_distortion == pytest.approx(z))
    # Again, above is the same as:
    no_distortion = project_world_to_image_with_distortion_K_R_t(
        K, R, t, np.zeros((5, 1)), pts)
    assert np.all(no_distortion == pytest.approx(z))
    # ... and also:
    no_distortion = project_world_to_image_with_distortion_K_R_C(
        K, R, C_from_R_t(R, t), np.zeros((5, 1)), pts)
    assert np.all(no_distortion == pytest.approx(z))
    # ... it doesn't matter whether we use C_from_R_t or C_from_Rt:
    no_distortion = project_world_to_image_with_distortion_K_R_C(
        K, R, C_from_Rt(Rt), np.zeros((5, 1)), pts)
    assert np.all(no_distortion == pytest.approx(z))
    # Finally, they're all the same as:
    tmp = project_world_to_image_K_R_C(K, R, C_from_Rt(Rt), pts)
    assert np.all(tmp == z)

    P = P_from_K_R_t(K, R, t)
    H = get_groundplane_to_image_homography(P)
    assert H.shape[0] == 3 and H.shape[1] == 3
    assert np.all(H[:, 0] == P[:, 0])
    assert np.all(H[:, 1] == P[:, 1])
    assert np.all(H[:, 2] == P[:, 3])

    Hinv = np.linalg.inv(H)
    H2 = get_image_to_groundplane_homography(P)
    assert np.all(Hinv == H2)

    # Divide by zero (NumPy should take care of this)
    np.seterr(divide='ignore', invalid='ignore')
    x = apply_projection(np.random.rand(3, 4), np.zeros((4, 1)))
    assert np.all(np.isnan(x))

    # Normalization of image coordinates
    K[0, 1] = 0.05
    K[0, 2] = 77
    K[1, 2] = 99
    img_pts = np.random.rand(2, 100)
    n_pts = normalize_image_coordinates(K, img_pts)
    Kinv = np.linalg.inv(K)
    assert np.all(apply_transformation(Kinv, img_pts) == n_pts)

    n_pts_wo = normalize_image_coordinates_with_distortion(
        K, np.zeros((5, )), img_pts)
    assert np.all(n_pts == pytest.approx(n_pts_wo))


def test_P():
    I = np.eye(3, 3)
    R = np.random.rand(3, 3)
    t = np.random.rand(3, 1).reshape((3, ))
    P = P_from_K_R_t(I, R, t)
    assert P.shape[0] == 3 and P.shape[1] == 4
    assert np.all(P == np.column_stack((R, t)))
    K = np.random.rand(3, 3)
    P = P_from_K_R_t(K, R, t)
    assert P.shape[0] == 3 and P.shape[1] == 4
    Rt = np.column_stack((R, t))
    for r in range(P.shape[0]):
        for c in range(P.shape[1]):
            assert P[r, c] == pytest.approx(dot(K[r, :], Rt[:, c]))


def test_shift_points_along_viewing_rays():
    invalid_pts = [
        np.float32(3),
        np.float64([17]),
        np.ones((2, )),
        np.ones((4, 1)),
        np.ones((1, 3))
    ]
    for ip in invalid_pts:
        with pytest.raises(ValueError):
            shift_points_along_viewing_rays(ip, 10)

    pts = np.random.rand(3, 42) * 23
    pts_dh = apply_dehomogenization(pts)
    for d in (23.7, np.random.rand(1, pts.shape[1])):
        shifted = shift_points_along_viewing_rays(pts, d)
        expected_x = np.multiply(np.asarray(d), pts_dh[0, :])
        expected_y = np.multiply(np.asarray(d), pts_dh[1, :])
        assert np.all(expected_x == shifted[0, :])
        assert np.all(expected_y == shifted[1, :])
        assert np.all(shifted[2, :] == d)

    with pytest.raises(ValueError):
        shift_points_along_viewing_rays(pts, np.random.rand(1, pts.shape[1] - 1))
    with pytest.raises(ValueError):
        shift_points_along_viewing_rays(pts, np.random.rand(1, pts.shape[1] + 1))
    with pytest.raises(ValueError):
        shift_points_along_viewing_rays(pts, np.random.rand(2, 3))


def test_compare_rotation_matrices():
    assert compare_rotation_matrices(np.eye(3, 3), np.eye(3, 3)) == 0
    assert compare_rotation_matrices(np.eye(3, 3), rot3d(10, 0, 0)) == pytest.approx(np.deg2rad(10))
    assert compare_rotation_matrices(np.eye(3, 3), rot3d(0, 45, 0)) == pytest.approx(np.deg2rad(45))
    assert compare_rotation_matrices(np.eye(3, 3), rot3d(0, 0, -30)) == pytest.approx(np.deg2rad(30))
    assert compare_rotation_matrices(np.eye(3, 3), rot3d(45, 45, 0)) == pytest.approx(np.deg2rad(62.7994296))



import numpy as np
import os
import pytest
from vito.imutils import flip_layers, imread, imsave, apply_on_bboxes, \
    ndarray2memory_file, memory_file2ndarray, roi, crop, pad, rgb2gray, \
    grayscale, pixelate, gaussian_blur, set_to, ensure_c3, concat, \
    rotate90, rotate180, rotate270, noop
from vito.pyutils import safe_shell_output


def test_flip_layers():
    # Invalid input gracefully handled
    assert flip_layers(None) is None

    # Single channel image
    x = np.random.rand(3, 3)
    xf = flip_layers(x)
    assert np.all(x == xf)

    # rgb to bgr conversion
    x = np.random.rand(16, 32, 3)
    xf = flip_layers(x)
    for c in range(x.shape[2]):
        assert np.all(x[:, :, c] == xf[:, :, x.shape[2]-c-1])

    # rgbA should be flipped to bgrA
    x = np.random.rand(17, 5, 4)
    xf = flip_layers(x)
    for c in range(x.shape[2]):
        complement = 2-c if c < 3 else 3
        assert np.all(x[:, :, c] == xf[:, :, complement])


def test_imread():
    exdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples')
    assert imread(None) is None
    for fn in ['', 'a-non-existing.file']:
        with pytest.raises(FileNotFoundError):
            imread(fn)
    # Forget the keyword to be passed on to PIL
    with pytest.raises(ValueError):
        imread(os.path.join(exdir, 'flamingo.jpg'), 'RGB')
    # Load RGB JPG
    img = imread(os.path.join(exdir, 'flamingo.jpg'))
    assert img.ndim == 3 and img.shape[2] == 3
    assert img.dtype == np.uint8
    # Same image, but BGR
    img_flipped = imread(os.path.join(exdir, 'flamingo.jpg'), flip_channels=True)
    for c in range(img.shape[2]):
        complement = 2-c if c < 3 else 3
        assert np.all(img[:, :, c] == img_flipped[:, :, complement])
    # Same image, but monochrome
    img = imread(os.path.join(exdir, 'flamingo.jpg'), mode='L')
    assert img.ndim == 2 or img.shape[2] == 1
    assert img.dtype == np.uint8
    # Load 8-bit single-channel PNG
    img = imread(os.path.join(exdir, 'peaks.png'))
    assert img.ndim == 2 or img.shape[2] == 1
    assert img.dtype == np.uint8
    # ... now enforce loading it as RGB/BGR
    img = imread(os.path.join(exdir, 'peaks.png'), mode='RGB')
    assert img.ndim == 3 and img.shape[2] == 3
    assert img.dtype == np.uint8
    # ... all channels should contain the same information
    for c in [1, 2]:
        assert np.all(img[:, :, c] == img[:, :, 0])
    # Load 16-bit PNG
    img = imread(os.path.join(exdir, 'depth.png'))
    assert img.ndim == 2 or img.shape[2] == 1
    assert img.dtype == np.int32
    # Load 16-bit PNG, specifying mode manually:
    img = imread(os.path.join(exdir, 'depth.png'), mode='I')
    assert img.ndim == 2 or img.shape[2] == 1
    assert img.dtype == np.int32
    # Load 1-bit PNG (requires specifying the mode!)
    img = imread(os.path.join(exdir, 'space-invader.png'), mode='L')
    assert img.ndim == 2 or img.shape[2] == 1
    assert img.dtype == np.uint8


def test_imsave(tmp_path):
    # This test will only work on Unix-based test environments because we
    # use the 'file' command to ensure the stored file is correct.
    exdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples')
    out_fn = str(tmp_path / 'test.png')
    ##########################################################################
    # Test RGB
    img_in = imread(os.path.join(exdir, 'flamingo.jpg'))
    assert img_in.ndim == 3 and img_in.shape[2] == 3
    assert img_in.dtype == np.uint8
    # Save (lossless) and reload
    imsave(out_fn, img_in)
    _, finfo = safe_shell_output('file', out_fn)
    assert finfo.split(':')[1].strip() == 'PNG image data, 400 x 400, 8-bit/color RGB, non-interlaced'
    img_out = imread(out_fn)
    assert img_out.ndim == 3 and img_out.shape[2] == 3
    assert img_out.dtype == np.uint8
    assert np.all(img_in[:] == img_out[:])
    ##########################################################################
    # Test RGB with flipping channels
    img_in = imread(os.path.join(exdir, 'flamingo.jpg'))
    assert img_in.ndim == 3 and img_in.shape[2] == 3
    assert img_in.dtype == np.uint8
    # Save (lossless) and reload
    imsave(out_fn, img_in, flip_channels=True)
    _, finfo = safe_shell_output('file', out_fn)
    assert finfo.split(':')[1].strip() == 'PNG image data, 400 x 400, 8-bit/color RGB, non-interlaced'
    img_out = imread(out_fn)
    assert img_out.ndim == 3 and img_out.shape[2] == 3
    assert img_out.dtype == np.uint8
    for c in range(3):
        assert np.all(img_in[:, :, c] == img_out[:, :, 2-c])
    ##########################################################################
    # Test monochrome 8 bit
    img_in = imread(os.path.join(exdir, 'peaks.png'))
    assert img_in.ndim == 2 or img_in.shape[2] == 1
    assert img_in.dtype == np.uint8
    # Save (lossless) and reload
    imsave(out_fn, img_in)
    _, finfo = safe_shell_output('file', out_fn)
    assert finfo.split(':')[1].strip() == 'PNG image data, 256 x 256, 8-bit grayscale, non-interlaced'
    img_out = imread(out_fn)
    assert img_out.ndim == 2 or img_out.shape[2] == 1
    assert img_out.dtype == np.uint8
    assert np.all(img_in[:] == img_out[:])
    ##########################################################################
    # Test monochrome 16 bit (will be loaded as np.int32, using PIL's 'I' mode)
    img_in = imread(os.path.join(exdir, 'depth.png'))
    assert img_in.ndim == 2 or img_in.shape[2] == 1
    assert img_in.dtype == np.int32
    # Explicitly cast to uint16
    img_in = img_in.astype(np.uint16)
    # Save (lossless) and reload
    imsave(out_fn, img_in)
    _, finfo = safe_shell_output('file', out_fn)
    assert finfo.split(':')[1].strip() == 'PNG image data, 800 x 600, 16-bit grayscale, non-interlaced'
    img_out = imread(out_fn)
    assert img_out.ndim == 2 or img_out.shape[2] == 1
    # Loading, however, will still produce a int32 image.
    assert img_out.dtype == np.int32
    assert np.all(img_in[:] == img_out[:])
    ##########################################################################
    # Test 1-bit PNG (requires specifying the mode!)
    img_in = imread(os.path.join(exdir, 'space-invader.png'), mode='L').astype(bool)
    assert img_in.ndim == 2 or img_in.shape[2] == 1
    assert img_in.dtype == bool
    imsave(out_fn, img_in)
    _, finfo = safe_shell_output('file', out_fn)
    assert (finfo.split(':')[1].strip() == 'PNG image data, 200 x 149, 1-bit colormap, non-interlaced' or
            finfo.split(':')[1].strip() == 'PNG image data, 200 x 149, 1-bit grayscale, non-interlaced')
    img_out = imread(out_fn, mode='L').astype(bool)
    assert img_out.ndim == 2 or img_out.shape[2] == 1
    assert np.all(img_in[:] == img_out[:])


def test_apply_on_bboxes():
    # Single and multi-channel test data
    x1 = np.zeros((5, 5), dtype=np.uint8)
    x2 = np.zeros((5, 5, 1), dtype=np.int32)
    x3 = np.zeros((5, 5, 3), dtype=np.uint8)

    # Example boxes
    boxes = [
        (0, 0, 1, 2),
        (5, 5, 3, 2),   # outside image
        (2, 3, 10, 9),  # should be clipped
        (3, 0, 0, 0),   # invalid
        (3, 1, 5, 1)    # should be clipped
    ]

    # Expected results
    e255 = x3.copy()
    e255[0:2, 0, :] = 255
    e255[3:, 2:, :] = 255
    e255[1, 3:, :] = 255

    e42 = x3.copy()
    e42[0:2, 0, :] = 42
    e42[3:, 2:, :] = 42
    e42[1, 3:, :] = 42

    # Exemplary functions
    def _set(img, value):
        img[:] = value
        return img

    def _set255(img):
        return _set(img, 255)

    assert apply_on_bboxes(None, boxes, _set255) is None

    # No kwargs:
    r1 = apply_on_bboxes(x1, boxes, _set255).copy()
    assert np.all(r1 == e255[:, :, 0])
    assert r1.dtype == np.uint8

    r2 = apply_on_bboxes(x2, boxes, _set255)
    e = e255[:, :, 0]
    assert np.all(r2.reshape(e.shape) == e)
    assert r2.dtype == np.int32

    r3 = apply_on_bboxes(x3, boxes, _set255)
    assert np.all(r3 == e255)
    assert r3.dtype == np.uint8

    # With kwargs
    r1 = apply_on_bboxes(x1, boxes, _set, 42)
    assert np.all(r1 == e42[:, :, 0])
    assert r1.dtype == np.uint8

    r2 = apply_on_bboxes(x2, boxes, _set, 42)
    assert np.all(r2 == e42)
    assert r2.dtype == np.int32

    r3 = apply_on_bboxes(x3, boxes, _set, 42)
    assert np.all(r3 == e42)
    assert r3.dtype == np.uint8


def test_np2mem():
    assert memory_file2ndarray(None) is None
    assert ndarray2memory_file(None) is None

    shapes = [(20, 128), (32, 64, 3), (64, 32, 3)]
    for s in shapes:
        x = (255.0 * np.random.randint(0, 255, s)).astype(np.uint8)
        mfpng = ndarray2memory_file(x, format='png')
        y = memory_file2ndarray(mfpng)
        assert np.array_equal(x, y)
        # Encode as JPG, this should decrease quality - at least we're not
        # able to decode the original input data
        mfjpg = ndarray2memory_file(x, format='jpeg')
        y = memory_file2ndarray(mfjpg)
        assert not np.array_equal(x, y)
    # Newer PIL versions can also work with 1D data:
    x = np.random.randint(0, 255, 3).astype(np.uint8)
    mf1d = ndarray2memory_file(x)
    y = memory_file2ndarray(mf1d)
    assert y.shape[0] == 3 and y.shape[1] == 1
    assert y[0, 0] == x[0] and y[1, 0] == x[1] and y[2, 0] == x[2]


def test_roi():
    assert roi(np.zeros((3, 3)), None) is None
    assert roi(np.zeros((3, 3)), [1, 2, 3, None]) is None
    assert roi(None, [1, 2, 3, 4]) is None
    invalid = roi(np.zeros((3, 3), dtype=np.uint8), [1, 1, 0, 1])
    assert invalid.shape[0] == 1 and invalid.shape[1] == 0
    invalid = roi(np.zeros((3, 3), dtype=np.uint8), [1, 1, 1, 0])
    assert invalid.shape[0] == 0 and invalid.shape[1] == 1

    # Proper clipping of corners:
    x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.int32)
    clipped = roi(x, [-1, -3, 2, 4])
    assert clipped.shape[0] == 1 and clipped.shape[1] == 1
    assert clipped[0, 0] == x[0, 0]
    clipped = roi(x, [3, -2, 2, 3])
    assert clipped.shape[0] == 1 and clipped.shape[1] == 1
    assert clipped[0, 0] == x[0, 2]

    rects = [
        (2, 5, 1, 2),
        (20, 20, 100, 100),
        (-5, 2, 1, 2)
    ]
    for shape in [(10, 10), (10, 10, 2)]:
        x = np.random.rand(*shape)
        for rect in rects:
            roi_ = roi(x, rect)
            l, t, w, h = rect
            r, b = l + w, t + h
            l = max(0, min(x.shape[1]-1, l))
            r = max(0, min(x.shape[1], r))
            t = max(0, min(x.shape[0]-1, t))
            b = max(0, min(x.shape[0], b))
            if x.ndim == 2:
                expected = x[t:b, l:r]
            else:
                expected = x[t:b, l:r, :]
            assert np.all(expected[:] == roi_[:])
    # Ensure crop() is an alias to roi()
    assert crop == roi


def test_pad():
    assert pad(None, 3) is None
    data = np.random.randint(0, 255, (2, 3))
    with pytest.raises(ValueError):
        pad(data, 0)
    with pytest.raises(ValueError):
        pad(data, -17)
    padded = pad(data, 1, color=None)
    assert padded.shape[0] == data.shape[0] + 2
    assert padded.shape[1] == data.shape[1] + 2
    assert padded.ndim == 3 and padded.shape[2] == 4
    assert np.all(padded[:, 0] == 0) and np.all(padded[0, :] == 0)
    assert np.all(padded[1, 1:-1, 0] == data[0, :])
    with pytest.raises(RuntimeError):
        pad(np.dstack((data, data)), 1, None)
    # Provide valid color as scalar and tuple
    padded = pad(data, 2, color=3)
    assert padded.shape[0] == data.shape[0] + 4
    assert padded.shape[1] == data.shape[1] + 4
    assert padded.ndim == data.ndim or (padded.ndim == 3 and padded.shape[2] == 1)
    assert np.all(padded[:, 0] == 3) and np.all(padded[0, :] == 3)
    assert np.all(padded[2:-2, 2:-2, 0] == data)
    # ... tuple
    padded = pad(data, 8, color=(200, 200, 200))
    assert padded.shape[0] == data.shape[0] + 16
    assert padded.shape[1] == data.shape[1] + 16
    assert padded.ndim == data.ndim or (padded.ndim == 3 and padded.shape[2] == 1)
    assert np.all(padded[:, 0] == 200) and np.all(padded[0, :] == 200)
    assert np.all(padded[8:-8, 8:-8, 0] == data)
    # Multi-channel image
    data = np.random.randint(0, 255, (2, 3, 3))
    color = np.random.randint(0, 255, (3,))
    padded = pad(data, 1, color=color)
    assert padded.shape[0] == data.shape[0] + 2
    assert padded.shape[1] == data.shape[1] + 2
    assert padded.ndim == data.ndim
    for c in range(3):
        assert np.all(padded[:, 0, c] == color[c]) and np.all(padded[0, :, c] == color[c])
        assert np.all(padded[1:-1, 1:-1, c] == data[:, :, c])
    # ... 4-channel
    data = np.random.randint(0, 255, (2, 3, 4))
    color = np.random.randint(0, 255, (4,))
    padded = pad(data, 1, color=color)
    assert padded.shape[0] == data.shape[0] + 2
    assert padded.shape[1] == data.shape[1] + 2
    assert padded.ndim == data.ndim and padded.shape[2] == 4
    for c in range(3):
        assert np.all(padded[:, 0, c] == color[c]) and np.all(padded[0, :, c] == color[c])
        assert np.all(padded[1:-1, 1:-1, c] == data[:, :, c])


def test_rgb2gray():
    assert rgb2gray(None) is None
    # Single channel input v1
    x = np.random.randint(0, 255, size=(17, 23), dtype=np.uint8)
    g = rgb2gray(x)
    assert np.array_equal(x, g)
    # Single channel input v2
    x = np.random.randint(0, 255, size=(17, 23, 1), dtype=np.uint8)
    g = rgb2gray(x)
    assert np.array_equal(x, g)
    # Single channel float
    x = np.random.rand(17, 23).astype(np.float32)
    g = rgb2gray(x)
    assert np.array_equal(x, g)
    # Cannot convert dual-channel to gray:
    x = np.random.randint(0, 255, size=(15, 10, 2), dtype=np.uint8)
    with pytest.raises(ValueError):
        g = rgb2gray(x)
    # RGB(A) input (uint8 and float32)
    inputs = [np.zeros((2, 3, 3), dtype=np.uint8),
              np.zeros((5, 17, 4), dtype=np.uint8),
              np.zeros((23, 5, 3), dtype=np.float32),
              np.zeros((18, 12, 4), dtype=np.float32)]
    for x, scalar, cast in zip(inputs, [255.0, 255.0, 1.0, 1.0], [np.uint8, np.uint8, np.float32, np.float32]):
        frgb = [0.2989, 0.5870, 0.1140]
        fbgr = frgb[::-1]
        for c in range(3):
            y = x.copy()
            y[:, :, c] = scalar
            g = rgb2gray(y, False)
            assert g.ndim == 2 or (g.ndim == 3 and g.shape[2] == 1)
            assert np.all(g[:] == cast(frgb[c]*scalar))
            g = rgb2gray(y, True)
            assert np.all(g[:] == cast(fbgr[c]*scalar))
    # Ensure grayscale is an alias to rgb2gray
    assert grayscale == rgb2gray


def test_pixelate():
    assert pixelate(None) is None
    with pytest.raises(ValueError):
        pixelate(np.zeros((17, 12)), 0, -1)
    x = np.random.randint(0, 255, (3, 10), dtype=np.uint8)
    res = pixelate(x)
    assert res.shape == x.shape
    assert np.all(res[:, :5] == res[0, 0])
    assert np.all(res[:, 5:] == res[-1, -1])

    x = np.random.randint(0, 255, (3, 12), dtype=np.uint8)
    res = pixelate(x)
    assert res.shape == x.shape
    # If the image size is not a multiple of the block size,
    # the blocks will be slightly larger (to avoid having small
    # -- recognizable -- blocks at the border)
    assert np.all(res[:, :6] == res[0, 0])
    assert np.all(res[:, 6:] == res[-1, -1])


def test_gaussian_blur():
    assert gaussian_blur(None) is None
    exdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples')
    img = imread(os.path.join(exdir, 'flamingo.jpg'))
    res1 = gaussian_blur(img, 3)
    res2 = gaussian_blur(img, 15)
    assert res1.shape == img.shape
    assert res2.shape == img.shape
    # We use PIL's ImageFilter module, so just resort to a simple sanity
    # check, i.e. whether blurring the image with a larger kernel degrades
    # its visual quality more than using a smaller kernel.
    diff1 = np.sqrt(np.sum((img - res1)**2))
    diff2 = np.sqrt(np.sum((img - res2)**2))
    assert diff1 > 0
    assert diff2 > diff1


def test_set_to():
    assert set_to(None, 3) is None
    assert set_to(np.zeros((2, 3)), None) is None

    # Test scalar
    res = set_to(np.zeros((2,)), 13)
    assert np.all(res[:] == 13) and res.shape == (2,)
    res = set_to(np.zeros((17, 12), dtype=np.uint8), 200)
    assert np.all(res[:] == 200) and res.shape == (17, 12)
    res = set_to(np.zeros((17, 12, 4), dtype=np.int32), -123)
    assert np.all(res[:] == -123) and res.shape == (17, 12, 4)

    # Test tuple
    with pytest.raises(ValueError):
        set_to(np.zeros((2, 3)), ())
    res = set_to(np.zeros((2,)), (13, 74, 99))
    assert np.all(res[:] == 13) and res.shape == (2,)
    res = set_to(np.zeros((3, 5), dtype=np.uint8), (78,))
    assert np.all(res[:] == 78) and res.shape == (3, 5)
    with pytest.raises(ValueError):
        set_to(np.zeros((17, 12, 4), dtype=np.int32), (1, 2, 3))
    val = (-123, 77, -23, 2**10)
    res = set_to(np.zeros((5, 6, 4), dtype=np.int32), val)
    assert res.shape == (5, 6, 4)
    for ch in range(res.shape[2]):
        assert np.all(res[:, :, ch] == val[ch])


def test_ensure_c3():
    # Invalid inputs
    assert ensure_c3(None) is None
    for invalid in [np.zeros(17), np.ones((4, 3, 2)), np.zeros((2, 2, 5))]:
        with pytest.raises(ValueError):
            ensure_c3(invalid)
    # Grayscale image (2-dim)
    x = np.random.randint(0, 255, (20, 30))
    c3 = ensure_c3(x)
    assert c3.ndim == 3 and c3.shape[2] == 3
    for c in range(3):
        assert np.array_equal(x, c3[:, :, c])
    # Grayscale image (3-dim, 1-channel)
    x = np.random.randint(0, 255, (10, 5, 1))
    c3 = ensure_c3(x)
    assert c3.ndim == 3 and c3.shape[2] == 3
    for c in range(3):
        assert np.array_equal(x[:, :, 0], c3[:, :, c])
    # RGB(A) inputs
    for x in [np.random.randint(0, 255, (12, 23, 3)), np.random.randint(0, 255, (12, 23, 4))]:
        c3 = ensure_c3(x)
        assert c3.ndim == 3 and c3.shape[2] == 3
        assert np.array_equal(x[:, :, :3], c3)


def test_concat():
    # Invalid inputs
    x = np.zeros((3,5))
    y = np.ones((5,7))
    assert concat(None, y, True) is None
    assert concat(x, None, False) is None
    with pytest.raises(ValueError):
        concat(x, y, True)
    with pytest.raises(ValueError):
        concat(x, y, False)
    
    # ndim == 2, different dtype
    x = np.random.randint(0, 255, (5, 17), dtype=np.uint8)
    y = np.random.randint(0, 255, (5, 2), dtype=np.int32)
    c = concat(x, y, True)
    # The concatenation should have the "better" dtype of the two inputs
    assert c.dtype == y.dtype
    assert np.array_equal(x.astype(c.dtype), c[:, :x.shape[1]])
    assert np.array_equal(y, c[:, x.shape[1]:])

    # ndim == 2, same dtype
    y = np.random.randint(0, 255, (5, 2), dtype=x.dtype)
    c = concat(x, y, True)
    assert c.dtype == x.dtype
    assert np.array_equal(x, c[:, :x.shape[1]])
    assert np.array_equal(y, c[:, x.shape[1]:])

    # ndim == 2 vs ndim == 3, single channel
    x = np.random.randint(0, 255, (5, 2, 1), dtype=np.uint8)
    y = np.random.randint(0, 255, (3, 2), dtype=np.int32)
    c = concat(x, y, False)
    assert c.dtype == y.dtype
    assert c.ndim == 3 and c.shape[2] == 1
    assert np.array_equal(x.astype(c.dtype), c[:x.shape[0], :, :])
    assert np.array_equal(y[:, :], c[x.shape[0]:, :, 0])

    # ndim == 2 vs ndim == 3, single vs multi-channel (compatible)
    x = np.random.randint(0, 255, (5, 2), dtype=np.int32)
    y = np.random.randint(0, 255, (3, 2, 4), dtype=np.uint8)
    c = concat(x, y, False)
    assert c.dtype == x.dtype
    assert c.ndim == 3 and c.shape[2] == y.shape[2]
    for l in range(c.shape[2]):
        assert np.array_equal(x[:, :], c[:x.shape[0], :, l])
    assert np.array_equal(y[:, :, :], c[x.shape[0]:, :, :])

    # ndim == 3, multi-channel
    # --> Incompatible inputs
    x = np.random.randint(0, 255, (5, 2, 2), dtype=np.uint8)
    y = np.random.randint(0, 255, (3, 2, 3), dtype=np.int32)
    with pytest.raises(ValueError):
        concat(x, y, False)
    # --> same number of channels
    x = np.random.randint(0, 255, (6, 2, 3), dtype=np.uint64)
    y = np.random.randint(0, 255, (6, 9, 3), dtype=np.int32)
    c = concat(x, y, True)
    assert c.dtype == x.dtype
    assert np.array_equal(x, c[:, :x.shape[1], :])
    assert np.array_equal(y, c[:, x.shape[1]:, :])
    # --> different number of channels (but compatible)
    x = np.random.randint(0, 255, (6, 2, 1), dtype=np.uint64)
    y = np.random.randint(0, 255, (6, 9, 3), dtype=np.int32)
    c = concat(x, y, True)
    assert c.dtype == x.dtype
    for l in range(3):
        assert np.array_equal(x[:, :, 0], c[:, :x.shape[1], l])
    assert np.array_equal(y.astype(c.dtype), c[:, x.shape[1]:])
    # ... switch parameter order
    c = concat(y, x, True)
    assert c.dtype == x.dtype
    assert np.array_equal(y.astype(c.dtype), c[:, :y.shape[1]])
    for l in range(3):
        assert np.array_equal(x[:, :, 0], c[:, y.shape[1]:, l])


def test_noop():
    for x in [None, 1, 3.7, np.zeros((3, 2)), 'test']:
        if isinstance(x, np.ndarray):
            assert np.array_equal(noop(x), x)
        else:
            assert noop(x) == x


def test_rotation():
    # Invalid inputs
    assert rotate90(None) is None
    assert rotate180(None) is None
    assert rotate270(None) is None

    for x in [np.array([[1, 2], [3, 4]], dtype=np.int32),
              np.random.randint(0, 255, (2, 2, 1), dtype=np.uint8),
              np.random.randint(0, 255, (2, 2, 3), dtype=np.uint8),
              np.random.randint(0, 255, (2, 2, 3), dtype=np.int64)]:
        if x.ndim == 2:
            y = rotate90(x)
            assert y[0, 0] == x[0, 1] and \
                y[0, 1] == x[1, 1] and \
                y[1, 0] == x[0, 0] and \
                y[1, 1] == x[1, 0]

            y = rotate180(x)
            assert y[0, 0] == x[1, 1] and \
                y[0, 1] == x[1, 0] and \
                y[1, 0] == x[0, 1] and \
                y[1, 1] == x[0, 0]

            y = rotate270(x)
            assert y[0, 0] == x[1, 0] and \
                y[0, 1] == x[0, 0] and \
                y[1, 0] == x[1, 1] and \
                y[1, 1] == x[0, 1]
        else:
            for l in range(x.shape[2]):
                y = rotate90(x)
                assert y[0, 0, l] == x[0, 1, l] and \
                    y[0, 1, l] == x[1, 1, l] and \
                    y[1, 0, l] == x[0, 0, l] and \
                    y[1, 1, l] == x[1, 0, l]

                y = rotate180(x)
                assert y[0, 0, l] == x[1, 1, l] and \
                    y[0, 1, l] == x[1, 0, l] and \
                    y[1, 0, l] == x[0, 1, l] and \
                    y[1, 1, l] == x[0, 0, l]

                y = rotate270(x)
                assert y[0, 0, l] == x[1, 0, l] and \
                    y[0, 1, l] == x[0, 0, l] and \
                    y[1, 0, l] == x[1, 1, l] and \
                    y[1, 1, l] == x[0, 1, l]



# import numpy as np
from types import SimpleNamespace


def iou(bbox1, bbox2):
    """
    Computes the intersection over union for the two bounding boxes
    given in rect representation: [left, top, width, height].
    """
    if bbox1 is None or bbox2 is None:
        return 0.0
    if isinstance(bbox1, BoundingBox):
        bbox1 = bbox1.to_rect_repr()
    if isinstance(bbox2, BoundingBox):
        bbox2 = bbox2.to_rect_repr()
    left1, top1, width1, height1 = bbox1
    left2, top2, width2, height2 = bbox2
    right1 = left1 + width1
    right2 = left2 + width2
    bottom1 = top1 + height1
    bottom2 = top2 + height2
    # Compute the coordinates of the intersection rectangle
    left = max(left1, left2)
    top = max(top1, top2)
    right = min(right1, right2)
    bottom = min(bottom1, bottom2)
    # Abort early if we don't have a valid intersection rectangle
    if right < left or bottom < top:
        return 0.0
    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box. Right and bottom coordinates are exclusive.
    intersection_area = (right - left) * (bottom - top)
    # Compute the area of both AABBs
    bb1_area = width1 * height1
    bb2_area = width2 * height2
    # Area of the union
    union_area = float(bb1_area + bb2_area - intersection_area)
    if union_area > 0.0:
        return intersection_area / union_area
    return 0.0  # pragma: no cover


def filter_detection_classes(detections, class_filter):
    """
    Returns the filtered list of detections that satisfy the given 'class_filter'.

    detections:     A list of Detection instances.

    class_filter:   Either 1) a list of object class IDs or labels
                    or 2) a single object class ID or label.
    """
    return [d for d in detections if d.is_class(class_filter)]


class Size(SimpleNamespace):
    @classmethod
    def from_hw(cls, height, width):
        return cls(width, height)

    @classmethod
    def from_wh(cls, width, height):
        return cls(width, height)

    def __init__(self, width, height):
        super().__init__(width=width, height=height)

    def area(self):
        return self.width * self.height


class Detection(SimpleNamespace):
    """Axis-aligned bounding box with object class ID and detection confidence/score."""
    def __init__(self, class_id, bounding_box, score):
        super().__init__(class_id=class_id, bounding_box=bounding_box, score=score)

    def is_class(self, class_filter):
        """
        Checks if this instance satisfies the given 'class_filter'.

        class_filter:  Either 1) a list of object class IDs or labels
                        or 2) a single object class ID or label.
        """
        if isinstance(class_filter, list):
            return any([self.class_id == c for c in class_filter])
        else:
            return self.class_id == class_filter
    
    def scale(self, scale_x, scale_y=None):
        """Scale the bounding box."""
        self.bounding_box.scale(scale_x, scale_y)
        return self


class BoundingBox(SimpleNamespace):
    """Axis-aligned bounding box."""

    @classmethod
    def from_corner_repr(cls, corner_repr, img_size=None):
        """
        Returns a BoundingBox from the 4-element argument corner_repr = [xmin, ymin, xmax, ymax].
        If img_size is not None, the representation will be scaled by the image dimension,
        i.e. img_size.width and img_size.height.
        """
        xmin, ymin, xmax, ymax = corner_repr
        w, h = xmax - xmin, ymax - ymin
        return cls(left=xmin, top=ymin, width=w, height=h, img_size=img_size)

    @classmethod
    def from_minmax_repr(cls, minmax_repr, img_size=None):
        """
        Returns a BoundingBox from the 4-element argument minmax_repr = [xmin, xmax, ymin, ymax].
        If img_size is not None, the representation will be scaled by the image dimension,
        i.e. img_size.width and img_size.height.
        """
        xmin, xmax, ymin, ymax = minmax_repr
        w, h = xmax - xmin, ymax - ymin
        return cls(left=xmin, top=ymin, width=w, height=h, img_size=img_size)

    @classmethod
    def from_centroid_repr(cls, centroid_repr, img_size=None):
        """
        Returns a BoundingBox from the 4-element argument centroid_repr = [cx, cy, w, h].
        If img_size is not None, the representation will be scaled by the image dimension,
        i.e. img_size.width and img_size.height.
        """
        cx, cy, w, h = centroid_repr
        left, top = cx - w/2, cy - h/2
        return cls(left=left, top=top, width=w, height=h, img_size=img_size)

    @classmethod
    def from_rect_repr(cls, rect_repr, img_size=None):
        """
        Returns a BoundingBox from the 4-element argument rect_repr = [left, top, width, height].
        If img_size is not None, the representation will be scaled by the image dimension,
        i.e. img_size.width and img_size.height.
        """
        return cls(left=rect_repr[0], top=rect_repr[1], width=rect_repr[2], height=rect_repr[3], img_size=img_size)

    def __init__(self, left, top, width, height, img_size=None):
        if img_size is not None:
            left *= img_size.width
            width *= img_size.width
            top *= img_size.height
            height *= img_size.height
        super().__init__(left=left, top=top, width=width, height=height)

    def scale(self, scale_x, scale_y=None):
        if scale_y is None:
            scale_y = scale_x
        self.left *= scale_x
        self.width *= scale_x
        self.top *= scale_y
        self.height *= scale_y

    def to_corner_repr(self, img_size=None):
        """
        Returns [xmin, ymin, xmax, ymax] as absolute values.
        If img_size is not None, the coordinates will be scaled by the image
        dimension, i.e. in the range [0, 1].
        """
        if img_size is None:
            return [self.left, self.top, self.left + self.width, self.top + self.height]
        else:
            return [self.left / img_size.width, self.top / img_size.height,
                    (self.left + self.width) / img_size.width,
                    (self.top + self.height) / img_size.height]

    def to_minmax_repr(self, img_size=None):
        """
        Returns [xmin, xmax, ymin, ymax] as absolute values.
        If img_size is not None, values will be given relative to the (image) size.
        """
        corner = self.to_corner_repr(img_size=img_size)
        return [corner[0], corner[2], corner[1], corner[3]]

    def to_centroid_repr(self, img_size=None):
        """
        Returns [center_x, center_y, width, height] as absolute values.
        If img_size is not None, values will be given relative to the (image) size.
        """
        cx = self.left + self.width/2
        cy = self.top + self.height/2
        if img_size is None:
            return [cx, cy, self.width, self.height]
        else:
            return [cx / img_size.width, cy / img_size.height,
                    self.width / img_size.width, self.height / img_size.height]

    def to_rect_repr(self, img_size=None):
        """
        Returns [xmin, ymin, width, height] in absolute values.
        If img_size is not None, the representation will be given
        relative to the (image) size.
        """
        if img_size is None:
            return [self.left, self.top, self.width, self.height]
        else:
            return [self.left / img_size.width, self.top / img_size.height,
                    self.width / img_size.width, self.height / img_size.height]

    def iou(self, other):
        """Returns the intersection over union."""
        return iou(self.to_rect_repr(), other)

    def area(self):
        return self.width * self.height


# Object categories from MS COCO
CATEGORIES_COCO = {
    0: 'background',
    1: 'person',
    2: 'bicycle',
    3: 'car',
    4: 'motorcycle',
    5: 'airplane',
    6: 'bus',
    7: 'train',
    8: 'truck',
    9: 'boat',
    10: 'traffic light',
    11: 'fire hydrant',
    13: 'stop sign',
    14: 'parking meter',
    15: 'bench',
    16: 'bird',
    17: 'cat',
    18: 'dog',
    19: 'horse',
    20: 'sheep',
    21: 'cow',
    22: 'elephant',
    23: 'bear',
    24: 'zebra',
    25: 'giraffe',
    27: 'backpack',
    28: 'umbrella',
    31: 'handbag',
    32: 'tie',
    33: 'suitcase',
    34: 'frisbee',
    35: 'skis',
    36: 'snowboard',
    37: 'sports ball',
    38: 'kite',
    39: 'baseball bat',
    40: 'baseball glove',
    41: 'skateboard',
    42: 'surfboard',
    43: 'tennis racket',
    44: 'bottle',
    46: 'wine glass',
    47: 'cup',
    48: 'fork',
    49: 'knife',
    50: 'spoon',
    51: 'bowl',
    52: 'banana',
    53: 'apple',
    54: 'sandwich',
    55: 'orange',
    56: 'broccoli',
    57: 'carrot',
    58: 'hot dog',
    59: 'pizza',
    60: 'donut',
    61: 'cake',
    62: 'chair',
    63: 'couch',
    64: 'potted plant',
    65: 'bed',
    67: 'dining table',
    70: 'toilet',
    72: 'tv',
    73: 'laptop',
    74: 'mouse',
    75: 'remote',
    76: 'keyboard',
    77: 'cell phone',
    78: 'microwave',
    79: 'oven',
    80: 'toaster',
    81: 'sink',
    82: 'refrigerator',
    84: 'book',
    85: 'clock',
    86: 'vase',
    87: 'scissors',
    88: 'teddy bear',
    89: 'hair drier',
    90: 'toothbrush'
}


# Object categories from PASCAL VOC 2007-2012
CATEGORIES_VOC07 = {
    0: "background",
    1: "aeroplane",
    2: "bicycle",
    3: "bird",
    4: "boat",
    5: "bottle",
    6: "bus",
    7: "car",
    8: "cat",
    9: "chair",
    10: "cow",
    11: "diningtable",
    12: "dog",
    13: "horse",
    14: "motorbike",
    15: "person",
    16: "pottedplant",
    17: "sheep",
    18: "sofa",
    19: "train",
    20: "tvmonitor"
}


# TODO encapsulate coco/voc via
# LabelMapVOC = LabelMap.from_map(...)
class LabelMap(object):
    @classmethod
    def from_map(cls, label_map, name=None):
        """Creates a label map from a dictionary of { class_id : label }."""
        return cls(label_map, name)
    
    @classmethod
    def from_list(cls, label_list, name=None):
        """Creates a label map from a list of category labels."""
        lm = { i: lbl for i,lbl in enumerate(label_list) }
        return cls(lm, name)
    
    def __init__(self, label_map, name):
        self.label_map = label_map
        self.name = name
    
    def label(self, c):
        """Returns the class label (string) for the given class ID or Detection instance 'c'."""
        if isinstance(c, Detection):
            cid = c.class_id
        else:
            cid = c
        if cid in self.label_map:
            return self.label_map[cid]
        raise ValueError()

    def class_id(self, lbl):
        """Returns the class ID (integer) for the given label (string)."""
        if lbl is None:
            raise ValueError()
        return list(self.label_map.keys())[list(self.label_map.values()).index(lbl.lower())]


LabelMapVOC07 = LabelMap.from_map(CATEGORIES_VOC07)


LabelMapCOCO = LabelMap.from_map(CATEGORIES_COCO)



#!/usr/bin/env python
# coding=utf-8
"""Utilities you'll often need when working with images ;-)"""

# TODOs:
# * imread/imsave should use tifffile for tiff files - supports multi-band images

import io
import logging
import os
import numpy as np
from PIL import Image
from PIL import ImageFilter


# Use logging module (fallback to default print) to log imutils set up (e.g. OpenCV vs Pillow).
__log_fx = print if len(logging.getLogger().handlers) == 0 else logging.getLogger().info


def flip_layers(nparray):
    """
    Flip RGB to BGR image data (numpy ndarray).
    Also accepts rgbA/bgrA and single channel images without crashing.
    """
    if nparray is None:
        return None
    if len(nparray.shape) == 3:
        if nparray.shape[2] == 4:
            # We got xyzA, make zyxA
            return nparray[..., [2, 1, 0, 3]]
        else:
            return nparray[:, :, ::-1]
    return nparray


def ensure_c3(nparray):
    """
    Ensures that the output image has 3 channels.
    Valid inputs: monochrome, 3- and 4-channel images.
    """
    if nparray is None:
        return None
    if nparray.ndim == 2:
        return np.repeat(nparray[:, :, np.newaxis], 3, axis=2)
    if nparray.ndim == 3:
        if nparray.shape[2] == 1:
            return np.repeat(nparray[:, :, :], 3, axis=2)
        elif nparray.shape[2] == 3:
            return nparray
        elif nparray.shape[2] == 4:
            return nparray[:, :, :3]
    raise ValueError("Invalid input to ensure_c3, input must be a 1-, 3- or 4-channel image.")


def rgb2gray(nparray, is_bgr=False):
    """
    Convert RGB image to grayscale using L = 0.2989 R + 0.5870 G + 0.1140 B.
    """
    if nparray is None:
        return None
    if nparray.ndim == 2 or (nparray.ndim == 3 and nparray.shape[2] == 1):
        return nparray
    if nparray.ndim == 3 and nparray.shape[2] == 2:
        raise ValueError('Cannot convert a 2 channel input image to grayscale.')
    if is_bgr:
        return np.dot(nparray[..., :3], [0.1140, 0.5870, 0.2989]).astype(nparray.dtype)
    else:
        return np.dot(nparray[..., :3], [0.2989, 0.5870, 0.1140]).astype(nparray.dtype)


# Alias for convenience
grayscale = rgb2gray


try:
    # Try to load OpenCV (in case you installed it in your workspace)
    import cv2
    __log_fx("vito.imutils will use OpenCV to save images via 'imsave'.")  # pragma: no cover

    def imsave(filename, image, flip_channels=False):  # pragma: no cover
        """Store an image using OpenCV."""
        # To be compatible with the Pillow/PIL version (see below), we have to
        # invert the flip_channels flag.
        if not flip_channels:
            cv2.imwrite(filename, flip_layers(image))
        else:
            cv2.imwrite(filename, image)
except:
    __log_fx("vito.imutils will use Pillow to save images via 'imsave' (OpenCV could not be loaded).")  # pragma: no cover
    # Fall back to Pillow
    def imsave(filename, image, flip_channels=False):
        """Store an image using PIL/Pillow, optionally flipping layers, i.e. BGR -> RGB."""
        if flip_channels:
            im_np = flip_layers(image)
        else:
            im_np = image
        Image.fromarray(im_np).save(filename)


def imread(filename, flip_channels=False, **kwargs):
    """Load an image (using PIL) into a NumPy array.
    Multi-channel images are returned as RGB unless you set flip_channels=True.

    Optional kwargs will be passed on to PIL's Image.convert(). Thus, you can
    specify PIL's loading 'mode', e.g. 'RGB' for color, 'RGBA' for a
    transparent image and 'L' for grayscale.
    """
    if filename is None:
        return None
    if not os.path.exists(filename):
        raise FileNotFoundError('Image %s does not exist' % filename)
    if not isinstance(flip_channels, bool):
        raise ValueError('Parameter "flip_channels" must be boolean - you probably forgot the keyword.')
    # PIL loads 16-bit PNGs as np.int32 ndarray. If we need to support other
    # bit depths, we should look into using pypng, see documentation at
    # https://pythonhosted.org/pypng/ex.html#reading
    # For now, PIL provides all we need
    image = np.asarray(Image.open(filename).convert(**kwargs))

    if flip_channels:
        return flip_layers(image)
    else:
        return image


def ndarray2memory_file(np_data, format='png'):
    """Convert numpy (image) array to io.BytesIO stream.
    Useful to stream an image via sockets."""
    if np_data is None:
        return None
    img = Image.fromarray(np_data)
    img_memory_file = io.BytesIO()
    img.save(img_memory_file, format)
    return img_memory_file


def memory_file2ndarray(memfile):
    """Load an image stored in the given io.BytesIO memory file memfile."""
    if memfile is None:
        return None
    return np.asarray(Image.open(memfile))


def clip_rect_to_image(rect, img_width, img_height):
    """Ensure that the rectangle is within the image boundaries.
    Explicitly casts all entries to integer.
    :param rect: list/tuple (l,t,w,h)
    :param img_width: int
    :param img_height: int
    :return: (l, t, w, h)
    """
    l, t, w, h = int(rect[0]), int(rect[1]), int(rect[2]), int(rect[3])
    # Right/bottom bounds are exclusive
    r = max(0, min(img_width, l+w))
    b = max(0, min(img_height, t+h))
    # Left/top bounds are inclusive
    l = max(0, l)
    t = max(0, t)
    return (l, t, r-l, b-t)


def is_valid_bbox(rect):
    """Left/top must be >= 0, W/H must be > 0"""
    return rect[0] >= 0 and rect[1] >= 0 and rect[2] > 0 and rect[3] > 0


def apply_on_bboxes(image_np, bboxes, func, *args, **func_kwargs):
    """Applies the function func (which returns a modified image) on each
    bounding box. Takes care of proper clipping, roi extraction and copying
    back the results.
    :param image_np:    numpy ndarray, input image
    :param bboxes:      list of bounding boxes, i.e. [(l, t, w, h), (...)]
    :param func:        function handle to apply on each bounding box
    :param func_kwargs: optional kwargs passed on to the function
    :return: numpy ndarray, a copy of the input image after applying the given
             function on each (valid) bounding box
    """
    if image_np is None:
        return None
    # Ensure the image is writeable
    image_np = image_np.copy()
    # Prevent invalid memory access
    bboxes = [clip_rect_to_image(bb, image_np.shape[1], image_np.shape[0])
            for bb in bboxes]
    bboxes = [b for b in bboxes if is_valid_bbox(b)]
    for bb in bboxes:
        l, t, w, h = bb
        roi = image_np[t:t+h, l:l+w]
        image_np[t:t+h, l:l+w] = func(roi, *args, **func_kwargs)
    return image_np


def roi(image_np, rect):
    """Returns the cropped ROI, with rect = [l, t, w, h]."""
    if image_np is None or rect is None or any([r is None for r in rect]):
        return None
    l, t, w, h = rect
    r, b = l+w, t+h
    img_height, img_width = image_np.shape[0], image_np.shape[1]

    # Right/bottom bounds are exclusive; left/top are inclusive
    l = max(0, min(img_width-1, l))
    r = max(0, min(img_width, r))
    t = max(0, min(img_height-1, t))
    b = max(0, min(img_height, b))

    if len(image_np.shape) == 2:
        return image_np[t:b, l:r]
    return image_np[t:b, l:r, :]


# Alias for convenience
crop = roi


def pad(image_np, border, color=None):
    """Pad the image by 'border' pixels in each direction.

    :param image_np: HxWxC numpy ndarray.
    :param border: int > 0.
    :param color: If None, border will be transparent.
                  Otherwise, the border will be drawn in the
                  corresponding color (RGB/BGR tuple or scalar).
    :return: (H+2*border)x(W+2*border)xC or x4 numpy ndarray.
    """
    if image_np is None:
        return None
    if border < 1:
        raise ValueError("Border must be > 0")
    h, w = image_np.shape[:2]
    c = image_np.shape[2] if image_np.ndim == 3 else 1
    if color is None:
        # RGB/BGR+A output
        out = np.zeros((h + 2*border, w + 2*border, 4), dtype=image_np.dtype)
        mask = np.zeros((h, w), dtype=image_np.dtype)
        mask[border:-border, border:-border] = np.iinfo(image_np.dtype).max
        out[border:-border, border:-border, 3] = mask
        if c in [1, 3, 4]:
            for i in range(max(3, c)):
                out[border:-border, border:-border, i] = image_np if c == 1 else image_np[:, :, i]
            return out
        else:
            raise RuntimeError("Invalid input shape, only 1, 3 or 4 channel images are supported!")
    else:
        out = np.zeros((h + 2*border, w + 2*border, c), dtype=image_np.dtype)
        if isinstance(color, (list, tuple, np.ndarray)):
            for i in range(c):
                out[:, :, i] = 0 if i >= len(color) else color[i]
        else:
            out[:] = color
        for i in range(c):
            out[border:-border, border:-border, i] = image_np if c == 1 else image_np[:, :, i]
        return out


def pixelate(image_np, block_width=5, block_height=-1):
    """
    Pixelate the image into blocks of size WxH.
    If H is -1, then H=W. Otherwise, rectangular blocks will
    be used (unless W is negative, in which case W=H will be used).
    """
    if image_np is None:
        return None
    if block_width < 1 and block_height < 1:
        raise ValueError("Either block width or height must be >= 1")
    block_width = block_height if block_width < 1 else block_width
    block_height = block_width if block_height < 1 else block_height
    resized_height = max(1, image_np.shape[0] // block_height)
    resized_width = max(1, image_np.shape[1] // block_width)
    pixelized_roi = Image.fromarray(image_np) \
        .resize((resized_width, resized_height), resample=Image.NEAREST) \
        .resize((image_np.shape[1], image_np.shape[0]), resample=Image.NEAREST)
    return np.asarray(pixelized_roi)


def gaussian_blur(image_np, radius=15):
    """Blurs the image using a Gaussian kernel with the given radius."""
    if image_np is None:
        return None
    img = Image.fromarray(image_np).filter(
        ImageFilter.GaussianBlur(radius=radius))
    return np.asarray(img)


def set_to(image_np, value):
    """Sets the image pixels to the given value."""
    if image_np is None or value is None:
        return None
    if isinstance(value, tuple):
        channels = 1 if image_np.ndim < 3 else image_np.shape[2]
        if len(value) < channels:
            raise ValueError('Input value has less entries than image channels.')
        if channels == 1:
            image_np[:] = value[0]
        else:
            for ch in range(channels):
                image_np[:, :, ch] = value[ch]
    else:
        image_np[:] = value
    return image_np



def _make_stack_compatible(img1, img2, horizontal):
    """Returns two "compatible" images to be used for horizontal or vertical
    stacking, i.e. they'll have the correct size, same dtype and number of
    channels."""
    # Check for compatible sizes
    if horizontal and img1.shape[0] != img2.shape[0]:
        raise ValueError('Images must have the same height for horizontal concatenation.')
    if not horizontal and img1.shape[1] != img2.shape[1]:
        raise ValueError('Images must have the same width for vertical concatenation.')
    # Check data types
    if img1.dtype != img2.dtype:
        if img1.dtype.itemsize > img2.dtype.itemsize:
            img2 = img2.astype(img1.dtype)
        else:
            img1 = img1.astype(img2.dtype)
    # Same dimensionality
    if img1.ndim < img2.ndim:
        img1 = img1[:, :, np.newaxis]
    elif img2.ndim < img1.ndim:
        img2 = img2[:, :, np.newaxis]
    # Check layers
    if img1.ndim == 3 and img1.shape[2] != img2.shape[2]:
        if img1.shape[2] == 1:
            img1 = np.repeat(img1[:, :, :], img2.shape[2], axis=2)
        elif img2.shape[2] == 1:
            img2 = np.repeat(img2[:, :, :], img1.shape[2], axis=2)
        else:
            raise ValueError('If channels are different, one of the inputs must be single-channel.')
    return img1, img2


def hstack(img1, img2):
    """Horizontally concatenates the two given images."""
    if img1 is None or img2 is None:
        return None
    img1, img2 = _make_stack_compatible(img1, img2, True)
    return np.hstack((img1, img2))


def vstack(img1, img2):
    """Vertically concatenates the two given images."""
    if img1 is None or img2 is None:
        return None
    img1, img2 = _make_stack_compatible(img1, img2, False)
    return np.vstack((img1, img2))


def concat(img1, img2, horizontal=True):
    """Concatenates the two given images either horizontally or vertically."""
    if horizontal:
        return hstack(img1, img2)
    else:
        return vstack(img1, img2)


def rotate90(image_np):
    """Rotates the given image by 90 degrees counter-clockwise."""
    if image_np is None:
        return None
    return np.rot90(image_np)


def rotate180(image_np):
    """Rotates the given image by 180 degrees."""
    if image_np is None:
        return None
    return np.fliplr(np.flipud(image_np))


def rotate270(image_np):
    """Rotates the given image by 270 degrees counter-clockwise."""
    if image_np is None:
        return None
    return np.rot90(image_np, -1)


def noop(x):
    """No-operation/identity function."""
    return x



import pytest
import numpy as np
from vito.detection2d import BoundingBox, Detection, Size, iou, filter_detection_classes,\
    LabelMapVOC07, LabelMapCOCO, LabelMap


def test_size():
    sz1 = Size.from_hw(60, 30)
    sz2 = Size.from_wh(30, 60)
    assert sz1 == sz2
    assert sz1.area() == 1800
    assert sz1.width == 30
    assert sz1.height == 60


def test_bbox_init():
    bb = BoundingBox.from_corner_repr([0, 0, 2, 4])
    assert bb.width == 2
    assert bb.height == 4
    assert bb.left == 0
    assert bb.top == 0

    bb_corner = BoundingBox.from_corner_repr([17, 42, 123, 99])
    assert bb_corner.width == 106
    assert bb_corner.height == 57
    assert bb_corner.left == 17
    assert bb_corner.top == 42

    bb_minmax = BoundingBox.from_minmax_repr([17, 123, 42, 99])
    assert bb_corner == bb_minmax

    bb_centroid = BoundingBox.from_centroid_repr([17+53, 42+28.5, 106, 57])
    assert bb_corner == bb_centroid

    bb_rect = BoundingBox.from_rect_repr(bb_centroid.to_rect_repr())
    assert bb_corner == bb_rect


def bb_approx_equal(a, b):
    """Test if two bounding boxes are approximately equal (up to rounding precision)."""
    assert a.left == pytest.approx(b.left)
    assert a.top == pytest.approx(b.top)
    assert a.width == pytest.approx(b.width)
    assert a.height == pytest.approx(b.height)


def test_bbox_conversion():
    bb = BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4))
    assert BoundingBox.from_rect_repr(bb.to_rect_repr()) == bb
    assert BoundingBox.from_corner_repr(bb.to_corner_repr()) == bb
    assert BoundingBox.from_centroid_repr(bb.to_centroid_repr()) == bb
    assert BoundingBox.from_minmax_repr(bb.to_minmax_repr()) == bb

    sz = Size(bb.width, bb.height)
    bb_rel = bb.to_rect_repr(sz)
    assert bb_rel[2] == 1
    assert bb_rel[3] == 1

    bb = BoundingBox.from_rect_repr([10, 70, 50, 140])
    bb_rel = bb.to_rect_repr(Size(100, 210))
    assert bb_rel[0] == 0.1
    assert bb_rel[1] == pytest.approx(1/3)
    assert bb_rel[2] == 0.5
    assert bb_rel[3] == pytest.approx(2/3)

    bb_rel = bb.to_corner_repr(Size(200, 210))
    assert bb_rel[0] == 0.05  # xmin
    assert bb_rel[1] == pytest.approx(1/3) # ymin
    assert bb_rel[2] == 0.30  # xmax
    assert bb_rel[3] == 1  # ymax

    bb_rel = bb.to_minmax_repr(Size(100, 210))
    assert bb_rel[0] == 0.1  # xmin
    assert bb_rel[1] == 0.6 # xmax
    assert bb_rel[2] == pytest.approx(1/3)  # ymin
    assert bb_rel[3] == 1  # ymax

    bb_rel = bb.to_centroid_repr(Size(200, 210))
    assert bb_rel[0] == 0.175  # cx
    assert bb_rel[1] == pytest.approx(2/3)  # cy
    assert bb_rel[2] == 0.25  # w
    assert bb_rel[3] == pytest.approx(2/3)  # h

    with pytest.raises(AttributeError):
        bb_abs = BoundingBox.from_centroid_repr(bb_rel, [100, 200])
    bb_abs = BoundingBox.from_centroid_repr(bb_rel, Size(200, 210))
    bb_approx_equal(bb_abs, bb)


def test_bbox_iou():
    bb = BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4))
    # Ensure w/h > 0
    bb.width += 1
    bb.height += 1
    # Input checks
    assert iou(None, bb) == 0.0
    assert iou(bb, None) == 0.0
    assert iou(None, None) == 0.0
    assert bb.iou(None) == 0.0
    assert bb.iou(bb) == iou(bb, bb)
    assert bb.iou(bb) == 1.0
    bb1 = BoundingBox.from_rect_repr([-10, -5, 7, 3])
    assert bb.iou(bb1) == 0.0
    assert bb1.iou(bb) == 0.0
    assert iou(bb1, bb) == 0.0
    bb2 = BoundingBox.from_rect_repr([-3, -3, 1, 2])
    assert bb1.iou(bb2) == 0.0

    bb1 = BoundingBox.from_rect_repr([10, 20, 30, 40])
    bb2 = BoundingBox.from_rect_repr([25, 20, 15, 40])
    assert bb1.iou(bb2) == 0.5
    bb1.width = 45
    assert bb1.iou(bb2) == 1/3


def test_bbox_area():
    bb = BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4))
    assert bb.area() == bb.width * bb.height
    bb.width = 0
    bb.height = 1
    assert bb.area() == 0


def test_detection2d():
    d1 = Detection(2, BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5)
    d2 = Detection("person", BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 1e7)

    assert d1.class_id == 2
    assert d1.score == 0.5
    assert d2.class_id == 'person'
    assert d2.score == 1e7

    dets = [
        d1, d2,
        Detection(20, BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5),
        Detection(2, BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5),
        Detection("car", BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5),
        Detection("person", BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5),
        Detection("person", BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5)
    ]
    ls = filter_detection_classes(dets, 2)
    assert len(ls) == 2
    # Filtering should keep the original order
    assert ls[0] == d1

    ls = filter_detection_classes(dets, "car")
    assert len(ls) == 1
    assert ls[0].class_id == 'car'

    ls = filter_detection_classes(dets, ["person", 2])
    assert len(ls) == 5
    assert ls[0] == d1
    assert ls[1] == d2
    assert ls[2].class_id == 2
    assert ls[3].class_id == 'person'
    assert ls[4].class_id == 'person'

    ls = filter_detection_classes(dets, [])
    assert len(ls) == 0
    ls = filter_detection_classes(dets, -2)
    assert len(ls) == 0

    d1 = Detection(2, BoundingBox.from_rect_repr([17, 3, 40, 10]), 0.5)
    d2 = d1.scale(3)
    assert d1.class_id == 2
    assert d1.score == 0.5
    assert d1.bounding_box.left == 51
    assert d1.bounding_box.width == 120
    assert d1.bounding_box.top == 9
    assert d1.bounding_box.height == 30
    assert d2 == d1

    d1.scale(0.5, 2)
    assert d1.class_id == 2
    assert d1.score == 0.5
    assert d1.bounding_box.left == 25.5
    assert d1.bounding_box.width == 60
    assert d1.bounding_box.top == 18
    assert d1.bounding_box.height == 60


def test_coco_labels():
    with pytest.raises(ValueError):
        LabelMapCOCO.class_id('foobar')
    with pytest.raises(ValueError):
        LabelMapCOCO.class_id(None)
    assert LabelMapCOCO.class_id('Person') == 1
    assert LabelMapCOCO.class_id('toothbrush') == 90

    det = Detection(2, BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5)
    assert LabelMapCOCO.label(det) == 'bicycle'

    with pytest.raises(ValueError):
        LabelMapCOCO.label(-1)
    assert LabelMapCOCO.label(1) == 'person'
    assert LabelMapCOCO.label(2) == 'bicycle'
    assert LabelMapCOCO.label(90) == 'toothbrush'


def test_voc07_labels():
    lbl_list = LabelMapVOC07.label_map.values()
    lm = LabelMap.from_list(lbl_list)

    with pytest.raises(ValueError):
        lm.class_id('foobar')
    with pytest.raises(ValueError):
        lm.class_id(None)

    det = Detection(3, BoundingBox.from_rect_repr(np.random.randint(0, 1e6, 4)), 0.5)
    assert lm.label(det) == 'bird'

    assert lm.class_id('Person') == 15
    assert lm.class_id('background') == 0
    assert lm.class_id('cow') == 10

    with pytest.raises(ValueError):
        lm.label(-1)
    assert lm.label(1) == 'aeroplane'
    assert lm.label(10) == 'cow'
    assert lm.label(20) == 'tvmonitor'



import numpy as np
import os
import pytest
from vito.flowutils import floread, flosave, colorize_flow


def test_floread():
    exfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples', 'color_wheel.flo')
    eximg = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples', 'flamingo.jpg')
    assert floread(None) is None
    for fn in ['', 'a-non-existing.file']:
        with pytest.raises(FileNotFoundError):
            floread(fn)
    # Load image as flow file
    with pytest.raises(ValueError):
        floread(eximg)
    # Load example .flo
    flow = floread(exfile)
    assert flow.ndim == 3 and flow.shape[2] == 2
    assert flow.shape[0] == 151
    assert flow.shape[1] == 151
    assert flow.dtype == np.float32


def test_flosave(tmp_path):
    exfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples', 'color_wheel.flo')
    # Load example .flo
    flow_in = floread(exfile)
    assert flow_in.ndim == 3 and flow_in.shape[2] == 2
    assert flow_in.dtype == np.float32

    # Save and reload
    out_fn = str(tmp_path / 'test.flo')
    flosave(out_fn, flow_in)
    flow_out = floread(out_fn)
    assert flow_out.ndim == 3 and flow_out.shape[2] == 2
    assert flow_out.dtype == np.float32
    assert np.all(flow_in[:] == flow_out[:])

    # Try saving float64 (will be loaded as float32)
    flow_in64 = flow_in.astype(np.float64)
    assert flow_in64.dtype == np.float64
    flosave(out_fn, flow_in64)
    flow_out = floread(out_fn)
    assert flow_out.ndim == 3 and flow_out.shape[2] == 2
    assert flow_out.dtype == np.float32
    assert np.all(flow_in[:] == flow_out[:])

    # Try saving an array of invalid shape as flow
    invalid = [
        np.zeros((3,), dtype=np.int32),
        np.zeros((15, 1), dtype=np.uint8),
        np.ones((10, 2), dtype=np.float32),
        np.zeros((20, 30, 1), dtype=np.float32),
        np.ones((4, 7, 3), dtype=np.float64)
    ]
    for i in invalid:
        with pytest.raises(ValueError):
            flosave(out_fn, i)


def assert_color_equal(a, b):
    for c in range(3):
        assert a[c] == b[c]


def test_colorize():
    # Try invalid inputs
    invalid = [
        np.zeros((3,), dtype=np.int32),
        np.zeros((15, 1), dtype=np.uint8),
        np.ones((10, 2), dtype=np.float32),
        np.zeros((20, 30, 1), dtype=np.float32),
        np.ones((4, 7, 3), dtype=np.float64)
    ]
    for i in invalid:
        with pytest.raises(ValueError):
            colorize_flow(i)

    examples = [
        np.zeros((10, 10, 2), dtype=np.uint8),
        np.zeros((10, 10, 2), dtype=np.float32),
        np.zeros((10, 10, 2), dtype=np.float64)
    ]
    for ex in examples:
        cf = colorize_flow(ex, max_val=None)
        assert np.all(cf[:] == 255)

    cf = colorize_flow(np.ones((10, 10, 2), dtype=np.float32))
    assert np.all(cf[:, :, 0] == 255)
    assert np.all(cf[:, :, 1] == 114)
    assert np.all(cf[:, :, 2] == 0)
    cf = colorize_flow(np.ones((10, 10, 2), dtype=np.float32), return_rgb=False)
    assert np.all(cf[:, :, 2] == 255)
    assert np.all(cf[:, :, 1] == 114)
    assert np.all(cf[:, :, 0] == 0)
    cf = colorize_flow(-np.ones((10, 10, 2), dtype=np.float32))
    assert np.all(cf[:, :, 0] == 0)
    assert np.all(cf[:, :, 1] == 52)
    assert np.all(cf[:, :, 2] == 255)
    # max_val should clip values into [0, max_val]
    cf = colorize_flow(-np.ones((10, 10, 2), dtype=np.float32), max_val=0)
    assert np.all(cf[:] == 255)



#!/usr/bin/env python
# coding=utf-8
"""Geometry utilities related to projections, camera calibration, etc."""

import numpy as np
import numbers

from . import pyutils as pu

# Check, if we can use np.matmul()
__has_np_matmul = pu.compare_version_strings(np.version.version, '1.10.0') >= 0

"""Util to use the correct matrix multiplication across different NumPy versions."""
matmul = np.matmul if __has_np_matmul else np.dot


def dot(a, b):
    """Dot product of Dx1 vectors.
    :params a,b: np.array
    """
    las = len(a.shape)
    lbs = len(b.shape)
    if las != lbs:
        if las == 1:
            a = a.reshape((a.shape[0], 1))
        if lbs == 1:
            b = b.reshape((b.shape[0], 1))
        las = len(a.shape)
        lbs = len(b.shape)

    if las > 1 and lbs > 1:
        assert a.shape == b.shape
        assert a.shape[1] == 1, 'Currently we only support vector dot products'

        sum = 0.0
        for i in range(a.shape[0]):
            sum += a[i] * b[i]
        return sum
    else:
        return np.dot(a, b)


def apply_transformation(T, pts):
    """Returns T*coords.
    :param T: MxD transformation matrix
    :param pts: DxN or (D-1)xN data points to transform. If dimension is D-1,
        a homogeneous dimension will be added.
    """
    if pts.ndim == 0:
        raise ValueError('Input coordinates must be DxN or (D-1)xN, not scalar.')
    T = T.astype(np.float64)
    pts = pts.astype(np.float64)
    ndim = T.shape[1]
    # Add homogeneous coordinate if necessary
    if pts.shape[0] == ndim-1:
        if pts.ndim == 1:
            npts = 1
        else:
            npts = pts.shape[1]
        pts = np.row_stack((pts, np.ones((1, npts), np.float64)))

    # Dimension check
    if pts.shape[0] != ndim:
        raise ValueError('Dimensions do not match')

    return matmul(T, pts)


# Alias
transform = apply_transformation


def apply_projection(P, pts):
    """Computes P*pts and returns the result after dehomogenization.
    :param P: MxD projection matrix
    :param pts: DxN or (D-1)xN data points to project. If dimension is D-1, a
        homogeneous dimension will be added.
    :returns: (M-1)xN projected points after dehomogenization.
    """
    return apply_dehomogenization(apply_transformation(P, pts))


def apply_dehomogenization(pts):
    """Computes the dehomogeneous points, i.e. dividing by the last dimension.
    :param pts: DxN points
    :returns: (D-1)xN dehomogeneous points

    This will issue a RuntimeWarning if the last dimension is 0. The
    corresponding division result will be set to nan.
    If you're aware of this, you can suppress the warnings, see:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.seterr.html
    """
    return np.divide(pts[:-1, :], pts[-1, :])


def shift_points_along_viewing_rays(pts, distance):
    """Shifts points along the viewing rays in a pinhole camera model. Usually
    used with 3xN pts: x and y coordinates will be computed, such that
    z=distance; distance can be a scalar or a 1xN array
    """
    if pts.ndim != 2 or pts.shape[0] != 3:
        raise ValueError('Input coordinates must be of shape 3xN.')
    pts = pts.astype(np.float64)
    num_pts = pts.shape[1]
    # Leverage similar triangles
    if isinstance(distance, numbers.Number):
        shifted = np.row_stack((pts[:-1, :] / pts[-1, :] * distance,
            np.array([distance]*num_pts, dtype=np.float64)))
    else:
        if distance.shape[1] != num_pts:
            raise ValueError('Number of pts and distances must match: {} points, but {} distances!'.format(
                num_pts, distance.shape[1]))
        distance = distance.astype(np.float64)
        shifted = np.row_stack((np.multiply(pts[:-1, :] / pts[-1, :], distance),
            distance))
    return shifted


def P_from_K_R_t(K, R, t):
    """Returns the 3x4 projection matrix P = K [R | t]."""
    K = K.astype(np.float64)
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    return matmul(K, np.column_stack((R, t)))


def t_from_R_C(R, C):
    """Returns the 3x1 translation vector t = -RC."""
    return -matmul(R, C)


def C_from_R_t(R, t):
    """Returns the origin of the coordinate system, C = -R't."""
    return -matmul(np.transpose(R), t)


def C_from_Rt(Rt):
    """Returns the origin of the coordinate system, C = -R't."""
    return C_from_R_t(Rt[:, :3], Rt[:, 3])


def Rt_from_R_C(R, C):
    """Returns the 3x4 extrinsics [R | t], where t = -RC."""
    return np.column_stack((R, t_from_R_C(R, C)))


def project_world_to_image_K_Rt(K, Rt, world_pts):
    """Projects 3D world points onto the image plane, assuming a calibrated
    camera and rectified inputs.
    """
    K = K.astype(np.float64)
    Rt = Rt.astype(np.float64)
    P = matmul(K, Rt)
    return apply_projection(P, world_pts)


def project_world_to_image_K_R_t(K, R, t, world_pts):
    """Project 3D world points onto the image plane, assuming you already
    calibrated your cameras and rectified your images.

    Don't confuse C (3D position vector of the camera's optical center) and
    t = -RC !!! If you have C, you'll want to use the other
    project_world_to_image().
    """
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    Rt = np.column_stack((R, t))
    return project_world_to_image_K_Rt(K, Rt, world_pts)


def project_world_to_image_K_R_C(K, R, C, world_pts):
    """Project 3D world points onto the image plane, assuming you already
    calibrated your cameras and rectified your images.

    Don't confuse C (3D position vector of the camera's optical center) and
    t = -RC !!! If you have t, you'll want to use
    project_world_to_image_K_R_t().
    """
    R = R.astype(np.float64)
    C = C.astype(np.float64)
    Rt = Rt_from_R_C(R, C)
    return project_world_to_image_K_Rt(K, Rt, world_pts)


def project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion.
    Note that Rt = [R, -RC], where C is the position vector of the optical
    center w.r.t. the world/reference coordinate system! Thus, t=-RC.

    See http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
    for notes on the camera model, how the distortion works, etc.
    """
    K = K.astype(np.float64)
    Rt = Rt.astype(np.float64)
    dist_coeff = dist_coeff.astype(np.float64)
    coords = coords.astype(np.float64)

    # Transform points from the world/reference coordinate system to the camera
    # coordinate system:
    coords_cam = apply_transformation(Rt, coords)

    # Perform the "normalized pinhole projection", i.e. shift the points along
    # the viewing rays to the "image plane". This is done via similar triangles:
    coords_normalized = coords_cam[0:2, :] / coords_cam[2, :]

    # Apply distortion model, since we use imperfect optics:
    kappa1, kappa2, rho1, rho2, kappa3 = dist_coeff
    r2 = np.sum(np.power(coords_normalized, 2), axis=0)  # squared radius, r^2 = x^2 + y^2
    r4 = np.multiply(r2, r2)
    r6 = np.multiply(r2, r4)

    rd = kappa1 * r2 + kappa2 * r4 + kappa3 * r6
    radial_distortion = np.multiply(coords_normalized, np.row_stack((rd, rd)))
    xy = np.multiply(coords_normalized[0, :], coords_normalized[1, :])
    x2 = np.power(coords_normalized[0, :], 2)
    y2 = np.power(coords_normalized[1, :], 2)
    tdx = 2.0 * rho1 * xy + rho2 * (r2 + 2.0 * x2)
    tdy = rho1 * (r2 + 2.0 * y2) + 2.0 * rho2 * xy
    tangential_distortion = np.row_stack((tdx, tdy))
    coords_corrected = coords_normalized + radial_distortion + tangential_distortion

    # Finally, project onto image:
    return apply_projection(K, coords_corrected)


# Alias for general projection
project = apply_projection


def project_world_to_image_with_distortion_K_R_C(K, R, C, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion. See project_world_to_image_with_distortion_K_Rt()
    """
    R = R.astype(np.float64)
    C = C.astype(np.float64)
    Rt = np.column_stack((R, -matmul(R, C)))
    return project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords)


def project_world_to_image_with_distortion_K_R_t(K, R, t, dist_coeff, coords):
    """Project 3D world points onto the image plane, compensating for the lens
    distortion. See project_world_to_image_with_distortion_K_Rt()
    """
    R = R.astype(np.float64)
    t = t.astype(np.float64)
    Rt = np.column_stack((R, t))
    return project_world_to_image_with_distortion_K_Rt(K, Rt, dist_coeff, coords)


def normalize_image_coordinates(K, coords):
    """Compute the normalized coordinates of the given pixel coordinates."""
    invK = np.linalg.inv(K.astype(np.float64))
    return apply_transformation(invK, coords)


def normalize_image_coordinates_with_distortion(K, dist_coeff, pixel_coords):
    """Computes the normalized pinhole projection coordinates of the given pixel
    coordinates and compensates for the lens distortion.
    Basically, it computes:
      x_dist = K^-1 * pixel_coords
      x_norm = compensate_distortion(x_dist)
    """
    # There's no closed form (or even nice) algebraic inverse for the distorted
    # projection. This code is ported from Bouguet's MATLAB toolbox normalize.m
    # We also use his variable names and naming conventions
    pixel_coords = pixel_coords.astype(np.float64)
    K = K.astype(np.float64)
    # In Bouguet's toolbox, K[0,:] = [f_x, alpha_c * f_x, cc_x],
    # alpha_c is the skew coefficient
    alpha_c = K[0, 1] / K[0, 0]
    fc = (K[0, 0], K[1, 1])
    cc = (K[0, 2], K[1, 2])
    dist_coeff = dist_coeff.astype(np.float64)

    # Subtract principal point, and divide by the focal length:
    xdx = (pixel_coords[0, :] - cc[0]) / fc[0]
    xdy = (pixel_coords[1, :] - cc[1]) / fc[1]

    # Second, undo skew
    xdx = xdx - alpha_c * xdy
    x_distort = np.row_stack((xdx, xdy))

    # Third, compensate for lens distortion:
    xn = compute_distortion_oulu(x_distort, dist_coeff)
    npts = xn.shape[1]
    return np.row_stack((xn, np.ones((1, npts), dtype=np.float64)))


def compute_distortion_oulu(xd, k):  # pragma: no cover
    """Compensates for radial and tangential distortion. Model From Oulu
    university. This code is a Python port from Bouguet's toolbox, namely
    comp_distortion_oulu.m. We use the same/similar variable names."""
    k1 = k[0]
    k2 = k[1]
    k3 = k[4]
    p1 = k[2]
    p2 = k[3]

    # Initial guess
    x = xd
    num_pts = xd.shape[1]
    iteration = 0
    while iteration < 20:
        r_2 = np.sum(np.power(x, 2), axis=0)
        # r_2 = sum(x.^2);
        k_radial = np.reshape(1.0 + k1 * r_2 + k2 * np.power(r_2, 2) + k3 * np.power(r_2, 3), (1, num_pts))
        _xy = np.multiply(x[0, :], x[1, :])
        dxx = 2.0 * p1 * _xy + p2 * (r_2 + 2.0 * np.power(x[0, :], 2))
        dxy = p1 * (r_2 + 2.0 * np.power(x[1, :], 2)) + 2.0 * p2 * _xy
        delta_x = np.row_stack((dxx, dxy))

        # Compute relative change (to enable early termination)
        magnitude_delta2 = np.sum(np.power(delta_x, 2), axis=0)
        magnitude_x2 = np.sum(np.power(x, 2), axis=0)
        change = np.divide(magnitude_delta2, magnitude_x2)

        x = np.divide(xd - delta_x, np.row_stack((k_radial, k_radial)))
        # Stop early if there is no major improvement:
        if np.max(change) < 1e-5:
            break
        iteration += 1
    return x


def get_groundplane_to_image_homography(P):
    """Given the 3x4 camera projection matrix P, returns the homography
    mapping ground plane points onto the image plane."""
    P = P.astype(np.float64)
    return P[:, [0, 1, 3]]


def get_image_to_groundplane_homography(P):
    """Given the 3x4 camera projection matrix P, returns the homography
    mapping image plane points onto the ground plane."""
    return np.linalg.inv(get_groundplane_to_image_homography(P))


def rotx3d(theta):
    """3D rotation matrix, x-axis, angle in radians."""
    ct = np.cos(theta)
    st = np.sin(theta)
    return np.array([
        [1.0, 0.0, 0.0],
        [0.0, ct, -st],
        [0.0, st, ct]], dtype=np.float64)


def roty3d(theta):
    """3D rotation matrix, y-axis, angle in radians."""
    ct = np.cos(theta)
    st = np.sin(theta)
    return np.array([
        [ct, 0.0, st],
        [0.0, 1.0, 0.0],
        [-st, 0.0, ct]], dtype=np.float64)


def rotz3d(theta):
    """3D rotation matrix, z-axis, angle in radians."""
    ct = np.cos(theta)
    st = np.sin(theta)
    return np.array([
        [ct, -st, 0.0],
        [st, ct, 0.0],
        [0.0, 0.0, 1.0]], dtype=np.float64)


def rot3d(deg_x, deg_y, deg_z):
    """Returns the 3D rotation matrix in ZYX (i.e. yaw-pitch-roll) order."""
    Rx = rotx3d(np.deg2rad(deg_x))
    Ry = roty3d(np.deg2rad(deg_y))
    Rz = rotz3d(np.deg2rad(deg_z))
    R = matmul(Rx, matmul(Ry, Rz))
    return R


def compare_rotation_matrices(R1, R2):
    """Returns the rotation angle (radians) between two 3x3 rotation matrices."""
    # Compute rotation matrix R_1-->2 as R1' * R2
    r_12 = matmul(np.transpose(R1), R2)
    # Compute the axis-angle representation using
    # trace(R_12) = 1 + 2*cos(theta) and return the
    # angle as rotation error/deviation.
    # max(-1, min(1, ...) prevents nan values due to
    # floating point precision.
    return np.arccos(max(-1, min(1, (np.trace(r_12) - 1.0) / 2.0)))



import numpy as np
import pytest
from vito.imvis import pseudocolor, color_by_id, exemplary_colors, overlay
from vito import colormaps


def assert_color_equal(a, b, flip=False):
    for c in range(3):
        if flip:
            assert a[c] == b[2-c]
        else:
            assert a[c] == b[c]


def test_pseudocolor():
    assert pseudocolor(None) is None

    data = np.array([[0, 1, 2], [255, 42, 0]], dtype=np.uint8)
    for cm in [colormaps.colormap_parula_rgb, colormaps.colormap_jet_rgb,
               colormaps.colormap_magma_rgb, colormaps.colormap_gray]:
        pc = pseudocolor(data, limits=None, color_map=cm)
        assert np.all(pc[0, 0, :] == pc[1, 2, :])
        assert_color_equal(pc[0, 1, :], cm[1])
        assert_color_equal(pc[0, 2, :], cm[2])
        assert_color_equal(pc[1, 0, :], cm[-1])
        assert_color_equal(pc[1, 1, :], cm[42])
        assert_color_equal(pc[1, 2, :], cm[0])

        pc = pseudocolor(data, limits=[0, 0], color_map=cm)
        for r in range(data.shape[0]):
            for c in range(data.shape[1]):
                assert_color_equal(pc[r, c, :], cm[0])

    data = np.random.rand(3, 7, 2)
    with pytest.raises(ValueError):
        _ = pseudocolor(data)

    data = np.array([0, 255], dtype=np.uint8)
    cm = colormaps.colormap_hot_bgr
    pc = pseudocolor(data, limits=[0, 255], color_map=cm)
    assert pc.shape == (2, 1, 3)
    assert_color_equal(pc[0, 0, :], cm[0])
    assert_color_equal(pc[1, 0, :], cm[-1])

    data = np.zeros((2, 3, 1), dtype=np.uint8)
    data[0, 0] = 20
    data[1, 2] = 10
    pc = pseudocolor(data, limits=None)
    cm = colormaps.colormap_parula_rgb
    assert_color_equal(pc[0, 0, :], cm[-1])
    assert_color_equal(pc[1, 0, :], cm[0])
    assert_color_equal(pc[1, 2, :], cm[127])

    # Test clipping
    pc = pseudocolor(data, limits=[2, 4])
    assert_color_equal(pc[0, 0, :], cm[-1])
    assert_color_equal(pc[1, 0, :], cm[0])
    assert_color_equal(pc[1, 2, :], cm[-1])

    pc = pseudocolor(data, limits=[30, 40])
    assert_color_equal(pc[0, 0, :], cm[0])
    assert_color_equal(pc[1, 0, :], cm[0])
    assert_color_equal(pc[1, 2, :], cm[0])

    # Test clipping by providing only a single value
    pc = pseudocolor(data, limits=[None, 4])
    assert_color_equal(pc[0, 0, :], cm[-1])
    assert_color_equal(pc[1, 0, :], cm[0])
    assert_color_equal(pc[1, 2, :], cm[-1])

    pc = pseudocolor(data, limits=[20, None])
    assert_color_equal(pc[0, 0, :], cm[0])
    assert_color_equal(pc[1, 0, :], cm[0])
    assert_color_equal(pc[1, 2, :], cm[0])

    # Test invalid inputs
    assert pseudocolor(np.array([np.nan, np.inf, np.NINF]), limits=None) is None
    pc = pseudocolor(np.array([np.nan, 255, 0]), limits=None, color_map=cm)
    assert pc.shape == (3, 1, 3)
    assert_color_equal(pc[0, 0, :], cm[0])
    assert_color_equal(pc[1, 0, :], cm[-1])
    assert_color_equal(pc[2, 0, :], cm[0])

    pc = pseudocolor(np.array([np.inf, 1024, 188]), limits=None, color_map=cm)
    assert pc.shape == (3, 1, 3)
    assert_color_equal(pc[0, 0, :], cm[0])
    assert_color_equal(pc[1, 0, :], cm[-1])
    assert_color_equal(pc[2, 0, :], cm[0])


def test_colormap_by_name():
    cm = colormaps.by_name('HSV')
    assert cm == colormaps.colormap_hsv_rgb
    cm = colormaps.by_name('HSV', return_rgb=False)
    assert cm == colormaps.colormap_hsv_bgr

    cm = colormaps.by_name('turbo')
    assert cm == colormaps.colormap_turbo_rgb
    cm = colormaps.by_name('TURBO', return_rgb=False)
    assert cm == colormaps.colormap_turbo_bgr

    cm = colormaps.by_name('gray')
    assert cm == colormaps.colormap_gray
    cm = colormaps.by_name('grayscale', return_rgb=False)
    assert cm == colormaps.colormap_gray
    cm = colormaps.by_name('grey')
    assert cm == colormaps.colormap_gray
    cm = colormaps.by_name('greyscale', return_rgb=False)
    assert cm == colormaps.colormap_gray

    assert colormaps.by_name(None) is None

    with pytest.raises(KeyError):
        colormaps.by_name('foo')


def test_sample_colormap():
    with pytest.raises(KeyError):
        colormaps.sample('Bar', 3)

    with pytest.raises(ValueError):
        colormaps.sample('Turbo', -1)

    with pytest.raises(ValueError):
        colormaps.sample('Turbo', 1)

    with pytest.raises(ValueError):
        colors = colormaps.sample('turbo', 0)

    colors = colormaps.sample('Turbo', 3)
    assert len(colors) == 3
    # Check all colors
    cm = colormaps.colormap_turbo_rgb
    assert_color_equal(colors[0], cm[0])
    assert_color_equal(colors[1], cm[127])
    assert_color_equal(colors[2], cm[255])

    colors = colormaps.sample('temperature', 4, return_rgb=False)
    assert len(colors) == 4
    cm = colormaps.colormap_temperature_bgr
    assert_color_equal(colors[0], cm[0])
    assert_color_equal(colors[1], cm[85])
    assert_color_equal(colors[2], cm[170])
    assert_color_equal(colors[3], cm[255])

    colors = colormaps.sample('HSV', 256)
    assert len(colors) == 256
    assert all([colors[i] == colormaps.colormap_hsv_rgb[i] for i in range(len(colors))])

    def check_oversampled(O, L, CM):
        assert len(O) == L
        num_reps = int(L / len(CM))
        for reps in range(num_reps):
            for e in range(len(CM)):
                assert_color_equal(O[reps*len(CM)+e], CM[e])
        num_rem = L % len(CM)
        for e in range(num_rem):
            assert_color_equal(O[num_reps*len(CM)+e], CM[e])

    oversampled = colormaps.sample('turbo', 1024)
    cm = colormaps.colormap_turbo_rgb
    check_oversampled(oversampled, 1024, cm)

    oversampled = colormaps.sample('turbo', 1025, return_rgb=False)
    cm = colormaps.colormap_turbo_bgr
    check_oversampled(oversampled, 1025, cm)

    oversampled = colormaps.sample('turbo', 751, return_rgb=False)
    cm = colormaps.colormap_turbo_bgr
    check_oversampled(oversampled, 751, cm)


def test_color_by_id():
    assert_color_equal(color_by_id(0), exemplary_colors[0])
    nc = len(exemplary_colors)
    assert_color_equal(color_by_id(nc), exemplary_colors[0])
    assert_color_equal(color_by_id(nc-1), exemplary_colors[-1])
    assert_color_equal(color_by_id(-1), exemplary_colors[-1])
    assert_color_equal(color_by_id(-3), exemplary_colors[-3])
    assert_color_equal(color_by_id(nc-1, flip_channels=True), exemplary_colors[-1], flip=True)
    assert_color_equal(color_by_id(-3, flip_channels=True), exemplary_colors[-3], flip=True)


def test_overlay():
    img1 = np.zeros((3, 3), dtype=np.uint8)
    img2 = 255 * np.ones((3, 3), dtype=np.uint8)
    with pytest.raises(ValueError):
        _ = overlay(img1, -1, img2)
    with pytest.raises(ValueError):
        _ = overlay(img1, 1.1, img2)

    # Invalid/Incompatible layers
    with pytest.raises(ValueError):
        _ = overlay(np.zeros((3, 3, 2)), 0.5, np.zeros((3, 3, 1)))

    # Overlay same channels (but not 1- or 3-channel)
    out = overlay(np.zeros((3, 3, 2)), 0.5, np.ones((3, 3, 2)))
    assert np.all(out[:] == pytest.approx(0.5))
    assert out.ndim == 3 and out.shape[2] == 2

    # Overlay gray on gray
    out = overlay(img1, 1, img2)
    assert np.all(img1[:] == out[:])
    out = overlay(img1, 0, img2)
    assert np.all(img2[:] == out[:])

    # Overlay gray on color
    img1 = np.zeros((3, 3, 3), dtype=np.uint8)
    out = overlay(img1, 1, img2)
    assert np.all(img1[:] == out[:])
    out = overlay(img2, 1, img1)
    assert np.all(img2[:] == out[:])
    img2 = 255 * np.ones((3, 3, 1), dtype=np.uint8)
    out = overlay(img1, 0, img2)
    assert np.all(img2[:] == out[:])
    out = overlay(img2, 1, img1)
    assert np.all(img2[:] == out[:])

    # Overlay color on color
    img2 = 255 * np.ones((3, 3, 3), dtype=np.uint8)
    out = overlay(img1, 0, img2)
    assert np.all(img2[:] == out[:])
    out = overlay(img1, 0.8, img2)
    assert np.all(50 == out[:])
    out = overlay(img1, 0.2, img2)
    assert np.all(204 == out[:])

    # Test different data type combination
    data_types = [np.uint8, np.float32, np.float64]
    expected_vals = {
        np.uint8: {
            np.uint8: 152,
            np.float32: pytest.approx(101.8),
            np.float64: pytest.approx(101.8)
        },
        np.float32: {
           np.uint8: 50,
           np.float32: pytest.approx(152.6),
           np.float64: pytest.approx(152.6)
        },
        np.float64: {
           np.uint8: 50,
           np.float32: pytest.approx(152.6),
           np.float64: pytest.approx(152.6)
        }
    }
    for dt1 in data_types:
        img1 = 255 * np.ones((10, 10, 3), dtype=dt1)
        for dt2 in data_types:
            img2 = 127 * np.ones((10, 10, 1), dtype=dt2)
            out = overlay(img1, 0.2, img2)
            assert np.all(out[:] == expected_vals[dt1][dt2])
            assert out.dtype == dt2

    # Unsupported dtypes
    with pytest.raises(ValueError):
        _ = overlay(np.zeros((3, 3), dtype=np.int32), 0.7, img1)
    with pytest.raises(ValueError):
        _ = overlay(img1, 0.2, np.zeros((3, 3), dtype=np.int32))

    # Test masked overlay
    # Size mismatch
    with pytest.raises(ValueError):
        _ = overlay(img1, 0.1, img1, np.ones((img1.shape[0]+1, img1.shape[1])))
    # Dimensionality mismatch
    with pytest.raises(ValueError):
        _ = overlay(img1, 0.1, img1, np.ones((img1.shape[0], img1.shape[1], 3)))
    for imshape in [(2, 3), (2, 3, 1), (3, 4, 3)]:
        img1 = np.ones(imshape, dtype=np.uint8)
        if img1.ndim > 2:
            img1[:, :, 0] = 50
        img2 = 2 * img1
        for dt in [np.uint8, np.float32]:
            for mask in [np.zeros((imshape[0], imshape[1]), dtype=dt),
                         np.zeros((imshape[0], imshape[1], 1), dtype=dt)]:
                for maskval in [1, 255]:
                    mask[1, 1] = maskval
                    out = overlay(img1, 0.2, img2, mask)
                    expval = np.uint8(255 * (0.2 * (img1[1, 1] / 255.0)\
                                      + 0.8 * (img2[1, 1] / 255.0)))
                    # Unmasked regions should be first image
                    assert np.all(out[:, 0] == img1[:, 0])
                    assert np.all(out[:, 2] == img1[:, 2])
                    assert np.all(out[0, :] == img1[0, :])
                    # Only masked regions should be blended
                    assert np.all(out[1, 1] == expval)
